/**
 * Copyright (C) 2013-2017 52°North Initiative for Geospatial Open Source
 * Software GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as publishedby the Free
 * Software Foundation.
 *
 * If the program is linked with libraries which are licensed under one of the
 * following licenses, the combination of the program with the linked library is
 * not considered a "derivative work" of the program:
 *
 *     - Apache License, version 2.0
 *     - Apache Software License, version 1.0
 *     - GNU Lesser General Public License, version 3
 *     - Mozilla Public License, versions 1.0, 1.1 and 2.0
 *     - Common Development and Distribution License (CDDL), version 1.0
 *
 * Therefore the distribution of the program linked with libraries licensed under
 * the aforementioned licenses, is permitted by the copyright holders if the
 * distribution is compliant with both the GNU General Public License version 2
 * and the aforementioned licenses.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.
 */
package org.n52.web;

import java.io.InputStream;

import org.n52.io.IoParameters;
import org.n52.io.v1.data.UndesignedParameterSet;
import org.n52.sensorweb.v1.spi.ParameterService;
import org.n52.sensorweb.v1.spi.RawDataService;

/**
 * Adapts SPI exceptions to HTTP specified Web exceptions.
 *
 * @author Henning Bredel <h.bredel@52north.org>
 *
 * @param <T> the parameter type of the service to adapt execptions for.
 */
public class WebExceptionAdapter<T> implements ParameterService<T>, RawDataService {

    private final ParameterService<T> composedService;

    public WebExceptionAdapter(ParameterService<T> toCompose) {
        this.composedService = toCompose;
    }

    @Override
    public T[] getExpandedParameters(IoParameters query) {
        T[] parameters = composedService.getExpandedParameters(query);
        if (parameters == null) {
            throw new InternalServerException("SPI implementation did return null value!");
        }
        return parameters;
    }

    @Override
    public T[] getCondensedParameters(IoParameters query) {
        T[] parameters = composedService.getCondensedParameters(query);
        if (parameters == null) {
            throw new InternalServerException("SPI implementation did return null value!");
        }
        return parameters;
    }

    @Override
    public T[] getParameters(String[] items) {
        T[] parameters = composedService.getParameters(items);
        if (parameters == null) {
            throw new InternalServerException("SPI implementation did return null value!");
        }
        return parameters;
    }

    @Override
    public T[] getParameters(String[] items, IoParameters query) {
        T[] parameters = composedService.getParameters(items, query);
        if (parameters == null) {
            throw new InternalServerException("SPI implementation did return null value!");
        }
        return parameters;
    }

    @Override
    public T getParameter(String item) {
        T parameter = composedService.getParameter(item);
        if (parameter == null) {
            throw new ResourceNotFoundException("Resource with id '" + item + "' was not found.");
        }
        return parameter;
    }

    @Override
    public T getParameter(String item, IoParameters query) {
        T parameter = composedService.getParameter(item, query);
        if (parameter == null) {
            throw new ResourceNotFoundException("Resource with id '" + item + "' was not found.");
        }
        return parameter;
    }
    
    @Override
	public InputStream getRawData(String id, IoParameters query) {
		if (composedService instanceof RawDataService) {
			return ((RawDataService)composedService).getRawData(id, query);
		}
		return null;
	}

	@Override
	public InputStream getRawData(UndesignedParameterSet parameters) {
		if (composedService instanceof RawDataService) {
			return ((RawDataService)composedService).getRawData(parameters);
		}
		return null;
	}

}
